from copy import deepcopy
from typing import List, Tuple, Union

from pypika import (
    Case,
    terms,
)

from fireant.dataset.fields import (
    DataType,
    Field,
    is_metric_field,
)
from fireant.dataset.modifiers import (
    DimensionModifier,
    ResultSet,
)
from ..utils import (
    alias_selector,
    flatten,
)


def _definition_field_for_data_blending(
    target_dataset_definition: Union[Field, terms.Term],
    target_dataset_leaf_definition: terms.Field,
    definition: terms.Term,
) -> terms.Term:
    """
    When using data blending, the dataset table of the set filter needs to be re-mapped to the table in the
    target dataset (i.e. primary or secondary). The easiest way to do that is to select the field in the
    target dataset directly. Otherwise table not found issues would pop up when resolving the joins.

    :param target_dataset_definition: The definition for a field in the target dataset.
    :param target_dataset_leaf_definition: The leaf definition for a field in the target dataset.
                                           Given sometimes a fireant's Field might have nested fireant's
                                           Fields.
    :param definition: A definition that might have its sub-parts (e.g. term, left, right) replaced.
                       That's likely the case for Criterion sub-classes and so on.
    :return: A term sub-class to be used in place of the provided definition argument, when applicable.
    """
    if isinstance(definition, (terms.ValueWrapper,)):
        # Constant values can be returned as is.
        return definition

    if isinstance(definition, (terms.Field,)) and definition == target_dataset_leaf_definition:
        target_dataset_leaf_definition.replace_table(target_dataset_leaf_definition.table, definition.table)
        return target_dataset_definition

    # Function, ...
    if hasattr(definition, 'args'):
        definition.args = [
            _definition_field_for_data_blending(target_dataset_definition, target_dataset_leaf_definition, arg)
            for arg in definition.args
        ]

    # CustomFunction, ...
    if hasattr(definition, 'params'):
        definition.params = [
            _definition_field_for_data_blending(target_dataset_definition, target_dataset_leaf_definition, param)
            for param in definition.params
        ]

    # RangeCriterion, ContainsCriterion, Not, All,
    if hasattr(definition, 'term'):
        definition.term = _definition_field_for_data_blending(
            target_dataset_definition, target_dataset_leaf_definition, definition.term
        )

    # BasicCriterion, ComplexCriterion, ...
    if hasattr(definition, 'left'):
        definition.left = _definition_field_for_data_blending(
            target_dataset_definition, target_dataset_leaf_definition, definition.left
        )

    if hasattr(definition, 'right'):
        definition.right = _definition_field_for_data_blending(
            target_dataset_definition, target_dataset_leaf_definition, definition.right
        )

    # Case
    if hasattr(definition, '_cases'):
        definition._cases = [
            _definition_field_for_data_blending(target_dataset_definition, target_dataset_leaf_definition, case)
            for case in definition._cases
        ]

    if hasattr(definition, '_else'):
        definition._else = _definition_field_for_data_blending(
            target_dataset_definition, target_dataset_leaf_definition, definition._else
        )

    return definition


def _make_set_dimension(set_filter: Field, target_dataset: 'DataSet') -> Field:
    """
    Returns a new dimension that uses a CASE statement as its definition, in order to represent membership to a set,
    given the provided conditional.

    :param set_filter: A ResultSet instance.
    :param target_dataset: A DataSet instance, that will be used as the dataset for which the new dimension
                           will be applied to.
    :return: A new Field instance.
    """
    old_definition = set_filter.filter

    while hasattr(old_definition, 'definition'):
        old_definition = old_definition.definition

    old_definition = deepcopy(old_definition)
    old_definition_sql = old_definition.get_sql(quote_char="")

    set_dimension = deepcopy(set_filter.filter.field)

    is_metric = is_metric_field(set_dimension)

    # When using data blending, the dataset table of the set filter needs to be re-mapped to the table in the
    # target dataset (i.e. primary or secondary). Otherwise table not found issues would pop up when resolving
    # the joins.
    if target_dataset and not is_metric:
        target_dataset_definition = target_dataset.fields[set_dimension.alias].definition
        target_dataset_leaf_definition = target_dataset_definition

        while hasattr(target_dataset_leaf_definition, 'definition'):
            # Sometimes a fireant's Field can have nested fireant Fields.
            target_dataset_leaf_definition = target_dataset_leaf_definition.definition

        old_definition = _definition_field_for_data_blending(
            target_dataset_definition, target_dataset_leaf_definition, old_definition
        )

    set_term = set_filter.set_label
    complement_term = set_filter.complement_label

    if not set_term and not complement_term:
        set_term = "set({})".format(old_definition_sql)
        complement_term = "complement({})".format(old_definition_sql)

    if not set_filter.will_group_complement:
        complement_term = set_filter.filter.field.definition

    if is_metric or not set_filter.will_replace_referenced_dimension:
        # When keeping a referenced dimension, we name the set dimension with a custom alias, so as to have no
        # alias clashes. That prevents issues with rollups/share dimensions, given the original dimension
        # is maintained. Also, metrics need to have the same treatment, given that, unlike dimensions, they are
        # never replaced.
        set_dimension.alias = alias_selector("set({})".format(old_definition_sql))

    set_dimension.data_type = DataType.text
    set_dimension.label = "Set({})".format(old_definition_sql) if not set_filter.set_label else set_filter.set_label
    set_dimension.definition = Case().when(old_definition, set_term).else_(complement_term)

    # Necessary for set feature to work properly with data blending's field mapping.
    set_dimension.is_artificial = True

    return set_dimension


def _unwrap_field(field: Field) -> Field:
    if isinstance(field, DimensionModifier):
        return field.dimension

    return field


def _replace_field_if_needed(field: Field, fields_per_set_filter, target_dataset: 'DataSet') -> Tuple[Field]:
    set_filter = fields_per_set_filter.get(_unwrap_field(field))

    if not set_filter:
        return (field,)

    set_dimension = _make_set_dimension(set_filter, target_dataset)

    if isinstance(field, DimensionModifier):
        modified_set_dimension = deepcopy(field)
        modified_set_dimension.dimension = set_dimension
        set_dimension = modified_set_dimension

    if set_filter.will_replace_referenced_dimension and not is_metric_field(set_dimension):
        # Metrics should not be replaced.
        return (set_dimension,)
    else:
        return (set_dimension, field)


def apply_set_dimensions(fields, filters, target_dataset: 'DataSet') -> List[Field]:
    """
    A transformed list of Field instances, in case `ResultSet` instances are present among the provided filters.

    :param fields: A list of Field instances.
    :param filters: A list of Filter instances.
    :param target_dataset: A DataSet instance.
    :return: A list of Field instances.
    """
    set_filters = [fltr for fltr in filters if isinstance(fltr, ResultSet)]

    if not set_filters:
        return [*fields]

    fields_per_set_filter = {set_filter.field: set_filter for set_filter in set_filters}
    fields_that_are_not_selected = set(fields_per_set_filter.keys())

    fields_with_set_dimensions = []
    for dimension_or_metric in fields:
        fields_with_set_dimensions.append(
            _replace_field_if_needed(dimension_or_metric, fields_per_set_filter, target_dataset)
        )
        unwrapped_field = _unwrap_field(dimension_or_metric)
        fields_that_are_not_selected.discard(unwrapped_field)
    fields_with_set_dimensions = flatten(fields_with_set_dimensions)

    for dimension_or_metric in fields_that_are_not_selected:
        set_filter = fields_per_set_filter[dimension_or_metric]

        if target_dataset and set_filter.field.alias not in target_dataset.fields:
            continue

        set_dimension = _make_set_dimension(set_filter, target_dataset)
        fields_with_set_dimensions.append(set_dimension)

    return fields_with_set_dimensions


def omit_set_filters(filters):
    """
    Returns all filters except the ones that are `ResultSet` instances.

    :param filters: A list of Filter instances.
    :return: A list of Filter instances.
    """
    return [fltr for fltr in filters if not isinstance(fltr, ResultSet)]



from pypika import (
    Parameter,
    PostgreSQLQuery,
    Table,
    terms,
    Dialects,
)
from pypika.terms import Node
from pypika.utils import format_quotes

from .base import Database


class DateTrunc(terms.Function):
    """
    Wrapper for the PostgreSQL date_trunc function
    """

    def __init__(self, field, date_format, alias=None):
        super(DateTrunc, self).__init__('DATE_TRUNC', date_format, field, alias=alias)


class PostgreSQLTimestamp(Node):
    def __init__(self, timestamp):
        self.timestamp = timestamp

    def get_sql(self, secondary_quote_char: str = "'", **kwargs) -> str:
        formatted_timestamp = format_quotes(self.timestamp.isoformat(), secondary_quote_char)
        return f"TIMESTAMP {formatted_timestamp}"


class PostgresDateAdd(terms.Function):
    def __init__(self, field, date_part, interval):
        wrapped_field = self.wrap_constant(field, PostgreSQLTimestamp)
        if date_part == "quarter":
            date_part = "month"
            interval *= 3

        interval_term = terms.Interval(**{f'{str(date_part)}s': interval, 'dialect': Dialects.POSTGRESQL})
        super().__init__('DATEADD', wrapped_field, interval_term)

    def get_function_sql(self, **kwargs):
        return " + ".join(self.get_arg_sql(arg, **kwargs) for arg in self.args)


class PostgreSQLDatabase(Database):
    """
    PostgreSQL client that uses the psycopg module.
    """

    # The pypika query class to use for constructing queries
    query_cls = PostgreSQLQuery

    def __init__(self, host="localhost", port=5432, database=None, user=None, password=None, **kwargs):
        super().__init__(host, port, database, **kwargs)
        self.user = user
        self.password = password

    def connect(self):
        import psycopg2

        return psycopg2.connect(
            host=self.host,
            port=self.port,
            dbname=self.database,
            user=self.user,
            password=self.password,
        )

    def trunc_date(self, field, interval):
        return DateTrunc(field, str(interval))

    def date_add(self, field, date_part, interval):
        return PostgresDateAdd(field, date_part, interval)

    def get_column_definitions(self, schema, table, connection=None):
        columns = Table("columns", schema="information_schema")

        columns_query = (
            PostgreSQLQuery.from_(columns, immutable=False)
            .select(columns.column_name, columns.data_type)
            .where(columns.table_schema == Parameter('%(schema)s'))
            .where(columns.field("table_name") == Parameter('%(table)s'))
            .distinct()
            .orderby(columns.column_name)
        )

        return self.fetch(str(columns_query), parameters=dict(schema=schema, table=table), connection=connection)



from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from pypika import (
    Parameter,
    Table,
    functions as fn,
    terms,
)
from pypika.dialects import SnowflakeQuery

from .base import Database

try:
    from snowflake import connector as snowflake
except:
    pass

IGNORED_SCHEMAS = {'INFORMATION_SCHEMA'}


class Trunc(terms.Function):
    """
    Wrapper for TRUNC function for truncating dates.
    """

    def __init__(self, field, date_format, alias=None):
        super(Trunc, self).__init__('TRUNC', field, date_format, alias=alias)


class SnowflakeDatabase(Database):
    """
    Snowflake client.
    """

    # The pypika query class to use for constructing queries
    query_cls = SnowflakeQuery

    DATETIME_INTERVALS = {'hour': 'HH', 'day': 'DD', 'week': 'W', 'month': 'MM', 'quarter': 'Q', 'year': 'Y'}
    _private_key = None

    def __init__(
        self,
        user='snowflake',
        password=None,
        account='snowflake',
        database='snowflake',
        private_key_data=None,
        private_key_password=None,
        region=None,
        warehouse=None,
        **kwags,
    ):
        super(SnowflakeDatabase, self).__init__(database=database, **kwags)
        self.user = user
        self.password = password
        self.account = account
        self.private_key_data = private_key_data
        self.private_key_password = private_key_password
        self.region = region
        self.warehouse = warehouse

    def connect(self):
        import snowflake

        return snowflake.connector.connect(
            database=self.database,
            account=self.account,
            user=self.user,
            password=self.password,
            private_key=self._get_private_key(),
            region=self.region,
            warehouse=self.warehouse,
        )

    def trunc_date(self, field, interval):
        trunc_date_interval = self.DATETIME_INTERVALS.get(str(interval), 'DD')
        return Trunc(field, trunc_date_interval)

    def date_add(self, field, date_part, interval):
        return fn.TimestampAdd(str(date_part), interval, field)

    def _get_private_key(self):
        if self._private_key is None:
            self._private_key = self._load_private_key_data()

        return self._private_key

    def _load_private_key_data(self):
        if self.private_key_data is None:
            return None

        private_key_password = None if self.private_key_password is None else self.private_key_password.encode()

        pkey = serialization.load_pem_private_key(
            self.private_key_data.encode(), private_key_password, backend=default_backend()
        )

        return pkey.private_bytes(
            encoding=serialization.Encoding.DER,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )

    def get_column_definitions(self, schema, table, connection=None):
        columns = Table('COLUMNS', schema='INFORMATION_SCHEMA')

        columns_query = (
            SnowflakeQuery.from_(columns, immutable=False)
            .select(columns.COLUMN_NAME, columns.DATA_TYPE)
            .where(columns.TABLE_SCHEMA == Parameter('%(schema)s'))
            .where(columns.field('TABLE_NAME') == Parameter('%(table)s'))
            .distinct()
            .orderby(columns.column_name)
        )

        return self.fetch(str(columns_query), connection=connection, parameters=dict(schema=schema, table=table))



from datetime import datetime
from functools import partial
from typing import Collection, Dict, Union
from typing import Iterable, List, Sequence, Type

import pandas as pd
from pypika import (
    Query,
    enums,
    terms,
)
from pypika import Table, functions as fn
from pypika.queries import QueryBuilder
from pypika.terms import Function

from fireant.dataset.fields import Field
from fireant.dataset.filters import Filter
from fireant.dataset.joins import Join
from fireant.exceptions import QueryCancelled
from fireant.middleware.decorators import apply_middlewares, connection_middleware
from fireant.queries.finders import (
    find_totals_dimensions,
    find_and_group_references_for_dimensions,
    find_required_tables_to_join,
    find_joins_for_tables,
)
from fireant.queries.references import make_reference_dimensions, make_reference_metrics, make_reference_filters
from fireant.queries.special_cases import adjust_daterange_filter_for_rolling_window
from fireant.queries.totals_helper import adapt_for_totals_query
from fireant.utils import (
    alias_selector,
    flatten,
)
from fireant.dataset.intervals import DatetimeInterval


class Database(object):
    """
    This is a abstract base class used for interfacing with a database platform.
    """

    # The pypika query class to use for constructing queries
    query_cls = Query

    slow_query_log_min_seconds = 15

    def __init__(
        self,
        host=None,
        port=None,
        database=None,
        max_result_set_size=200000,
        middlewares=[],
    ):
        self.host = host
        self.port = port
        self.database = database
        self.max_result_set_size = max_result_set_size
        self.middlewares = middlewares + [connection_middleware]

    def connect(self):
        """
        This function must establish a connection to the database platform and return it.
        """
        raise NotImplementedError

    def cancel(self, connection):
        """
        Cancel any running query.
        """
        if hasattr(connection, "cancel"):
            connection.cancel()
        else:
            # A default cancel for databases for which no specific cancel is implemented
            # This will force an exit of the connection context manager
            raise QueryCancelled("Query was cancelled")

    def get_column_definitions(self, schema, table, connection=None):
        """
        Return a list of column name, column data type pairs.

        :param schema: The name of the table schema.
        :param table: The name of the table to get columns from.
        :param connection: (Optional) The connection to execute this query with.
        :return: A list of columns.
        """
        raise NotImplementedError

    def trunc_date(self, field, interval):
        """
        This function must create a Pypika function which truncates a Date or DateTime object to a specific interval.
        """
        raise NotImplementedError

    def date_add(self, field: terms.Term, date_part: str, interval: int):
        """
        This function must add/subtract a Date or Date/Time object.
        """
        raise NotImplementedError

    def convert_date(self, dt: datetime) -> Union[datetime, Function]:
        """
        Override to provide a custom function for converting a date.
        Defaults to an identity function.

        :param dt: Date to convert
        """
        return dt

    def to_char(self, definition):
        return fn.Cast(definition, enums.SqlTypes.VARCHAR)

    @apply_middlewares
    def fetch_queries(self, *queries, connection=None, parameters: Union[Dict, Collection] = ()):
        results = []
        # Parameters can either be passed as a list when using formatting placeholders like %s (varies per platform)
        # or a dict when using named placeholders.
        for query in queries:
            cursor = connection.cursor()
            cursor.execute(str(query), parameters)
            results.append(cursor.fetchall())

        return results

    def fetch(self, query, **kwargs):
        return self.fetch_queries(query, **kwargs)[0]

    @apply_middlewares
    def execute(self, *queries, **kwargs):
        connection = kwargs.get("connection")
        for query in queries:
            cursor = connection.cursor()
            cursor.execute(str(query))
            connection.commit()

    @apply_middlewares
    def fetch_dataframes(self, *queries, parse_dates=None, **kwargs):
        connection = kwargs.get("connection")
        dataframes = []
        for query in queries:
            dataframes.append(pd.read_sql(query, connection, coerce_float=True, parse_dates=parse_dates))
        return dataframes

    def fetch_dataframe(self, query, **kwargs):
        return self.fetch_dataframes(query, **kwargs)[0]

    def __str__(self):
        return f'Database|{self.__class__.__name__}|{self.host}'

    def make_slicer_query_with_totals_and_references(
        self,
        table,
        joins,
        dimensions,
        metrics,
        operations,
        filters,
        references,
        orders,
        share_dimensions=(),
    ) -> List[Type[QueryBuilder]]:
        """
        The following two loops will run over the spread of the two sets including a NULL value in each set:
         - reference group (WoW, MoM, etc.)
         - dimension with roll up/totals enabled (totals dimension)

        This will result in at least one query where the reference group and totals dimension is NULL, which shall be
        called base query. The base query will ALWAYS be present, even if there are zero reference groups or totals
        dimensions.

        For a concrete example, check the test case in :
        ```
        fireant.tests.queries.test_build_dimensions.QueryBuilderDimensionTotalsTests
            #test_build_query_with_totals_cat_dimension_with_references
        ```
        """

        filters = adjust_daterange_filter_for_rolling_window(dimensions, operations, filters)

        totals_dimensions = find_totals_dimensions(
            dimensions,
            share_dimensions,
        )
        totals_dimensions_and_none = [None] + totals_dimensions[::-1]

        reference_groups = find_and_group_references_for_dimensions(dimensions, references)
        reference_groups_and_none = [(None, None)] + list(reference_groups.items())

        queries = []
        for totals_dimension in totals_dimensions_and_none:
            (dimensions_with_totals, filters_with_totals) = adapt_for_totals_query(
                totals_dimension,
                dimensions,
                filters,
            )

            for reference_parts, references in reference_groups_and_none:
                (dimensions_with_ref, metrics_with_ref, filters_with_ref,) = self.adapt_for_reference_query(
                    reference_parts,
                    dimensions_with_totals,
                    metrics,
                    filters_with_totals,
                    references,
                )
                query = self.make_slicer_query(
                    table,
                    joins,
                    dimensions_with_ref,
                    metrics_with_ref,
                    filters_with_ref,
                    orders,
                )

                # Add these to the query instance so when the data frames are joined together, the correct references and
                # totals can be applied when combining the separate result set from each query.
                query._totals = totals_dimension
                query._references = references
                queries.append(query)

        return queries

    def adapt_for_reference_query(self, reference_parts, dimensions, metrics, filters, references):
        if reference_parts is None:
            return dimensions, metrics, filters

        ref_dim, unit, interval = reference_parts

        offset_func = partial(self.date_add, date_part=unit, interval=interval)
        offset_func_inv = partial(self.date_add, date_part=unit, interval=-interval)

        ref_dimensions = make_reference_dimensions(
            dimensions, ref_dim, offset_func, self.transform_field_to_query, self.trunc_date
        )
        ref_metrics = make_reference_metrics(metrics, references[0].reference_type.alias)
        ref_filters = make_reference_filters(filters, ref_dim, offset_func_inv)

        return ref_dimensions, ref_metrics, ref_filters

    def make_slicer_query(
        self,
        base_table: Table,
        joins: Sequence[Join] = (),
        dimensions: Sequence[Field] = (),
        metrics: Sequence[Field] = (),
        filters: Sequence[Filter] = (),
        orders: Sequence = (),
    ) -> Type[QueryBuilder]:
        """
        Creates a pypika/SQL query from a list of slicer elements.

        This is the base implementation shared by two implementations: the query to fetch data for a slicer request and
        the query to fetch choices for dimensions.

        This function only handles dimensions (select+group by) and filtering (where/having), which is everything needed
        for the query to fetch choices for dimensions.

        The slicer query extends this with metrics, references, and totals.

        :param base_table:
            pypika.Table - The base table of the query, the one in the FROM clause
        :param joins:
            A collection of joins available in the slicer. This should include all slicer joins. Only joins required for
            the query will be used.
        :param dimensions:
            A collection of dimensions to use in the query.
        :param metrics:
            A collection of metrics to use in the query.
        :param filters:
            A collection of filters to apply to the query.
        :param orders:
            A collection of orders as tuples of the metric/dimension to order by and the direction to order in.

        :return:
        """
        query = self.query_cls.from_(base_table, immutable=False)
        elements = flatten([metrics, dimensions, filters])

        # Add joins
        join_tables_needed_for_query = find_required_tables_to_join(elements, base_table)

        for join in find_joins_for_tables(joins, base_table, join_tables_needed_for_query):
            query = query.join(join.table, how=join.join_type).on(join.criterion)

        # Add dimensions
        for dimension in dimensions:
            dimension_term = self.transform_field_to_query(dimension, self.trunc_date)
            query = query.select(dimension_term)

            if dimension.groupable:
                query = query.groupby(dimension_term)

        # Add filters
        for fltr in filters:
            query = query.having(fltr.definition) if fltr.is_aggregate else query.where(fltr.definition)

        # Add metrics
        metric_terms = [self.transform_field_to_query(metric) for metric in metrics]
        if metric_terms:
            query = query.select(*metric_terms)

        # In the case that the orders are determined by a field that is not selected as a metric or dimension, then it needs
        # to be added to the query.
        select_aliases = {el.alias for el in query._selects}
        for (orderby_field, orientation) in orders:
            orderby_term = self.transform_field_to_query(orderby_field)
            query = query.orderby(orderby_term, order=orientation)

            if orderby_term.alias not in select_aliases:
                query = query.select(orderby_term)

        return query

    def make_latest_query(
        self,
        base_table: Table,
        joins: Iterable[Join] = (),
        dimensions: Iterable[Field] = (),
    ):
        query = self.query_cls.from_(base_table, immutable=False)

        # Add joins
        join_tables_needed_for_query = find_required_tables_to_join(dimensions, base_table)
        for join in find_joins_for_tables(joins, base_table, join_tables_needed_for_query):
            query = query.join(join.table, how=join.join_type).on(join.criterion)

        for dimension in dimensions:
            f_dimension_key = alias_selector(dimension.alias)
            query = query.select(fn.Max(dimension.definition).as_(f_dimension_key))

        return query

    def transform_field_to_query(self, field, window=None):
        """
        Makes a list of pypika terms for a given dataset field.

        :param field:
            A field from a dataset.
        :param window:
            A window function to apply to the dimension definition if it is a continuous dimension.
        :return:
            a list of terms required to select and group by in a SQL query given a dataset dimension. This list will contain
            either one or two elements. A second element will be included if the dimension has a definition for its display
            field.
        """
        f_alias = alias_selector(field.alias)

        if window and isinstance(field, DatetimeInterval):
            return window(field.definition, field.interval_key).as_(f_alias)

        return field.definition.as_(f_alias)



from pypika import (
    Dialects,
    MySQLQuery,
    Table,
    enums,
    functions as fn,
    terms,
)
from pypika.terms import CustomFunction, Interval, Parameter

from . import sql_types
from .base import Database
from .type_engine import TypeEngine
from ..exceptions import QueryCancelled

_DateFormat = CustomFunction('DATE_FORMAT', ['field', 'date_format'])
_DateSub = CustomFunction('DATE_SUB', ['field', 'interval'])
_MakeDate = CustomFunction('MAKEDATE', ['year', 'day_of_year'])
_Weekday = CustomFunction('WEEKDAY', ['arg'])
_Year = CustomFunction('YEAR', ['arg'])
_Quarter = CustomFunction('QUARTER', ['arg'])
_Timestamp = CustomFunction('TIMESTAMP', ['arg'])


class DateAdd(terms.Function):
    """
    Override for the MySQL specific DateAdd function which expects an interval instead of the date part and interval
    unit e.g. DATE_ADD("date", INTERVAL 1 YEAR)
    """

    def __init__(self, field, interval_term, alias=None):
        super(DateAdd, self).__init__('DATE_ADD', field, interval_term, alias=alias)


class _CustomMySQLInterval(terms.Function):
    """
    The PyPika Interval function generates Interval SQL based on the provided time unit arguments.
    Fireant needs a more flexible Interval query string generator than this
    to mimicking the Vertica TRUNC function, so this is defined locally here.
    """

    def __init__(self, field, unit):
        super().__init__('INTERVAL', field, unit)
        self.field = field
        self.unit = unit

    def get_sql(self, **kwargs):
        field_sql = self.field.get_sql(**kwargs)
        return f'INTERVAL {field_sql} {self.unit}'


class MySQLDatabase(Database):
    """
    MySQL client that uses the PyMySQL module.
    """

    # The pypika query class to use for constructing queries
    query_cls = MySQLQuery

    def __init__(
        self,
        host='localhost',
        port=3306,
        database=None,
        user=None,
        password=None,
        charset='utf8mb4',
        read_timeout=None,
        **kwargs,
    ):
        super(MySQLDatabase, self).__init__(host, port, database, **kwargs)
        self.user = user
        self.password = password
        self.charset = charset
        self.read_timeout = read_timeout
        self.type_engine = MySQLTypeEngine()

    def connect(self):
        """
        Returns a MySQL connection

        :return: pymysql Connection class
        """
        import pymysql

        return pymysql.connect(
            host=self.host,
            port=self.port,
            db=self.database,
            user=self.user,
            password=self.password,
            charset=self.charset,
            read_timeout=self.read_timeout,
            cursorclass=pymysql.cursors.Cursor,
        )

    def cancel(self, connection):
        try:
            connection.kill(connection.thread_id())
        except RuntimeError:
            # The kill operation seems to always throw a RuntimeError after sending the command while reading the
            # OK packet. Ignore this exception and raise our own QueryCancelled.
            # A bunch of other errors will be thrown as well because the socket has been set to None and calls are still
            # being made to the connector, but that's fine.
            pass
        finally:
            raise QueryCancelled("Query was cancelled")

    def trunc_date(self, field, interval):
        if interval == 'hour':
            return _DateFormat(field, '%Y-%m-%d %H:00:00')
        elif interval == 'day':
            return _DateFormat(field, '%Y-%m-%d 00:00:00')
        elif interval == 'week':
            return _DateFormat(_DateSub(field, _CustomMySQLInterval(_Weekday(field), 'DAY')), '%Y-%m-%d 00:00:00')
        elif interval == 'month':
            return _DateFormat(field, '%Y-%m-01')
        elif interval == 'quarter':
            return (
                _MakeDate(_Year(_Timestamp(field)), 1)
                + _CustomMySQLInterval(_Quarter(_Timestamp(field)), 'QUARTER')
                - Interval(quarters=1, dialect=Dialects.MYSQL)
            )
        elif interval == 'year':
            return _DateFormat(field, '%Y-01-01')
        else:
            raise ValueError(f'Invalid interval provided to trunc_date method: {interval}')

    def to_char(self, definition):
        return fn.Cast(definition, enums.SqlTypes.CHAR)

    def date_add(self, field, date_part, interval):
        # adding an extra 's' as MySQL's interval doesn't work with 'year', 'week' etc, it expects a plural
        interval_term = terms.Interval(**{'{}s'.format(str(date_part)): interval, 'dialect': Dialects.MYSQL})
        return DateAdd(field, interval_term)

    def get_column_definitions(self, schema, table, connection=None):
        columns = Table('columns', schema='INFORMATION_SCHEMA')

        columns_query = (
            MySQLQuery.from_(columns)
            .select(columns.column_name, columns.column_type)
            .where(columns.table_schema == Parameter('%(schema)s'))
            .where(columns.field('table_name') == Parameter('%(table)s'))
            .distinct()
            .orderby(columns.column_name)
        )

        return self.fetch(str(columns_query), parameters=dict(schema=schema, table=table), connection=connection)

    def import_csv(self, table, file_path, connection=None):
        """
        Execute a query to import a file into a table using the provided connection.

        :param table: The name of a table to import data into.
        :param file_path: The path of the file to be imported.
        :param connection: (Optional) The connection to execute this query with.
        """
        import_query = MySQLQuery.load(file_path).into(table)

        self.execute(str(import_query), connection=connection)

    def create_temporary_table_from_columns(self, table, columns, connection=None):
        """
        Creates a temporary table from a list of columns.

        :param table: The name of the new temporary table.
        :param columns: The columns of the new temporary table.
        :param connection: (Optional) The connection to execute this query with.
        """
        create_query = MySQLQuery.create_table(table).temporary().columns(*columns)

        self.execute(str(create_query), connection=connection)

    def create_temporary_table_from_select(self, table, select_query, connection=None):
        """
        Creates a temporary table from a SELECT query.

        :param table: The name of the new temporary table.
        :param select_query: The query to be used for selecting data of an existing table for the new temporary table.
        :param connection: (Optional) The connection to execute this query with.
        """
        create_query = MySQLQuery.create_table(table).temporary().as_select(select_query)

        self.execute(str(create_query), connection=connection)


class MySQLTypeEngine(TypeEngine):
    mysql_to_ansi_mapper = {
        'bit': sql_types.Char,
        'char': sql_types.Char,
        'nchar': sql_types.Char,
        'varchar': sql_types.VarChar,
        'nvarchar': sql_types.VarChar,
        'text': sql_types.Text,
        'boolean': sql_types.Boolean,
        'int': sql_types.Integer,
        'integer': sql_types.Integer,
        'year': sql_types.Integer,
        'smallint': sql_types.SmallInt,
        'tinyint': sql_types.SmallInt,
        'bigint': sql_types.BigInt,
        'decimal': sql_types.Decimal,
        'fixed': sql_types.Decimal,
        'numeric': sql_types.Numeric,
        'float': sql_types.Float,
        'real': sql_types.DoublePrecision,
        'double': sql_types.DoublePrecision,
        'date': sql_types.Date,
        'time': sql_types.Time,
        'datetime': sql_types.DateTime,
        'timestamp': sql_types.Timestamp,
    }

    ansi_to_mysql_mapper = {
        'CHAR': 'char',
        'VARCHAR': 'varchar',
        'TEXT': 'text',
        'BOOLEAN': 'boolean',
        'INTEGER': 'integer',
        'SMALLINT': 'smallint',
        'BIGINT': 'bigint',
        'DECIMAL': 'decimal',
        'NUMERIC': 'numeric',
        'FLOAT': 'float',
        'REAL': 'real',
        'DOUBLEPRECISION': 'real',
        'DATE': 'date',
        'TIME': 'time',
        'TIMESTAMP': 'timestamp',
    }

    def __init__(self):
        super(MySQLTypeEngine, self).__init__(self.mysql_to_ansi_mapper, self.ansi_to_mysql_mapper)



import logging
from typing import Iterable, List, Tuple, Type

import pandas as pd
from pypika.queries import QueryBuilder

from fireant.database import Database
from fireant.dataset.fields import DataType, Field
from fireant.dataset.modifiers import RollupValue
from fireant.dataset.references import calculate_delta_percent
from fireant.dataset.totals import get_totals_marker_for_dtype
from fireant.queries.finders import find_field_in_modified_field, find_totals_dimensions
from fireant.queries.pandas_workaround import df_subtract
from fireant.utils import alias_selector, chunks


logger = logging.getLogger(__name__)

# Passing in an empty dictionary as format option to pandas' parse_dates causes errors to be ignored instead of
# being coerced into NaT.
PANDAS_TO_DATETIME_FORMAT = {}


def fetch_data(
    database: Database,
    queries: List[Type[QueryBuilder]],
    dimensions: Iterable[Field],
    share_dimensions: Iterable[Field] = (),
    reference_groups=(),
) -> Tuple[int, pd.DataFrame]:
    queries = [str(query) for query in queries]

    # Indicate which dimensions need to be parsed as date types
    # For this we create a dictionary with the dimension alias as key and PANDAS_TO_DATETIME_FORMAT as value
    pandas_parse_dates = {}
    for dimension in dimensions:
        unmodified_dimension = find_field_in_modified_field(dimension)
        if unmodified_dimension.data_type == DataType.date:
            pandas_parse_dates[alias_selector(unmodified_dimension.alias)] = PANDAS_TO_DATETIME_FORMAT

    results = database.fetch_dataframes(*queries, parse_dates=pandas_parse_dates)
    max_rows_returned = 0
    for result_df in results:
        row_count = len(result_df)
        if row_count > max_rows_returned:
            max_rows_returned = row_count
        if row_count > database.max_result_set_size:
            logger.warning('row_count_over_max', extra={'row_count': len(result_df), 'database': str(database)})
            # drop all result rows above database.max_result_set_size in place
            result_df.drop(result_df.index[database.max_result_set_size :], inplace=True)

    logger.info('max_rows_returned', extra={'row_count': max_rows_returned, 'database': str(database)})
    return max_rows_returned, reduce_result_set(results, reference_groups, dimensions, share_dimensions)


def reduce_result_set(
    results: Iterable[pd.DataFrame],
    reference_groups,
    dimensions: Iterable[Field],
    share_dimensions: Iterable[Field],
):
    """
    Reduces the result sets from individual queries into a single data frame. This effectively joins sets of references
    and concatenates the sets of totals.

    :param results: A list of data frame
    :param reference_groups: A list of groups of references (grouped by interval such as WoW, etc)
    :param dimensions: A list of dimensions, used for setting the index on the result data frame.
    :param share_dimensions: A list of dimensions from which the totals are used for calculating share operations.
    :return:
    """
    # One result group for each rolled up dimension. Groups contain one member plus one for each reference type used.
    result_groups = chunks(results, 1 + len(reference_groups))

    dimension_keys = [alias_selector(d.alias) for d in dimensions]
    totals_dimension_keys = [alias_selector(d.alias) for d in find_totals_dimensions(dimensions, share_dimensions)]
    dimension_dtypes = result_groups[0][0][dimension_keys].dtypes

    # Reduce each group to one data frame per rolled up dimension
    group_data_frames = []
    for i, result_group in enumerate(result_groups):
        if dimension_keys:
            result_group = [result.set_index(dimension_keys) for result in result_group]

        base_df = result_group[0]
        reference_dfs = [
            _make_reference_data_frame(base_df, result, reference)
            for result, reference_group in zip(result_group[1:], reference_groups)
            for reference in reference_group
        ]

        # Merge the different reference dataframes into 1
        merged_df = base_df
        for reference_df in reference_dfs:
            merged_df = pd.merge(
                merged_df, reference_df, how="left", left_index=True, right_index=True, suffixes=('', '_delete')
            )
            # Drop duplicate columns from the merge (indicate by suffix _delete)
            merged_df.drop(merged_df.filter(regex='_delete$').columns.tolist(), axis=1, inplace=True)

        # If there are rolled up dimensions in this result set then replace the NaNs for that dimension value with a
        # marker to indicate totals.
        # The data frames will be ordered so that the first group will contain the data without any rolled up
        # dimensions, then followed by the groups with them, ordered by the last rollup dimension first.
        if totals_dimension_keys[:i]:
            merged_df = _replace_rollup_constants_for_totals_markers(merged_df, dimension_dtypes)

        group_data_frames.append(merged_df)

    return pd.concat(group_data_frames, sort=False).sort_index(na_position="first")


def _replace_rollup_constants_for_totals_markers(data_frame, dtypes):
    # some things are just easier to do without an index. Reset it temporarily to replace Rollup constants with the
    # rollup marker values
    index_names = data_frame.index.names
    data_frame.reset_index(inplace=True)

    for dimension_key, dtype in dtypes.items():
        data_frame[dimension_key] = data_frame[dimension_key].replace(
            RollupValue.CONSTANT, get_totals_marker_for_dtype(dtype)
        )

    return data_frame.set_index(index_names)


def _make_reference_data_frame(base_df, ref_df, reference):
    """
    This applies the reference metrics to the data frame given the base data frame and the reference data frame.

    When a reference is selected as a delta or a delta percentage, the calculation is performed here. Otherwise, the
    reference data frame is returned.

    :param base_df:
    :param ref_df:
    :param reference:
    :return:
    """
    metric_column_indices = [i for i, column in enumerate(ref_df.columns) if column not in base_df.columns]
    ref_columns = [ref_df.columns[i] for i in metric_column_indices]

    if not (reference.delta or reference.delta_percent):
        return ref_df[ref_columns]

    original_ref_df = ref_df

    base_columns = [base_df.columns[i] for i in metric_column_indices]

    # Select just the metric columns from the DF and rename them with the reference key as a suffix
    base_df, ref_df = base_df[base_columns].copy(), ref_df[ref_columns].copy()
    # Both data frame columns are renamed in order to perform the calculation below.
    base_df.columns = ref_df.columns = [
        column.replace(reference.reference_type.alias, reference.alias) for column in ref_columns
    ]

    ref_delta_df = df_subtract(base_df, ref_df, fill_value=0)

    if reference.delta_percent:
        ref_delta_df = calculate_delta_percent(ref_df, ref_delta_df)

    # Add original reference values back to df
    ref_delta_df[ref_columns] = original_ref_df[ref_columns]
    # Alternative for the above line but with copying:
    # ref_delta_df = ref_delta_df.join(original_ref_df[ref_columns])

    return ref_delta_df



import numpy as np

from typing import Callable, List
from pypika import functions as fn
from pypika.queries import QueryBuilder

from fireant import utils
from fireant.utils import immutable
from .fields import Field
from .modifiers import FieldModifier


class ReferenceFilter:
    def __init__(self, metric: Field, operator: str, value: int):
        self.metric = metric
        self.operator = operator
        self.value = value


class Reference(FieldModifier):
    def __init__(
        self,
        field: Field,
        reference_type: 'ReferenceType',
        delta: bool = False,
        delta_percent: bool = False,
        filters: List[ReferenceFilter] = (),
    ):
        super().__init__(field)

        self.reference_type = reference_type
        self.alias = (
            reference_type.alias + "_delta_percent"
            if delta_percent
            else reference_type.alias + "_delta"
            if delta
            else reference_type.alias
        )

        self.label = (
            reference_type.label + " Δ%"
            if delta_percent
            else reference_type.label + " Δ"
            if delta
            else reference_type.label
        )

        self.time_unit = reference_type.time_unit
        self.interval = reference_type.interval

        self.delta = delta_percent or delta
        self.delta_percent = delta_percent

        self.filters = filters

    def __eq__(self, other):
        return isinstance(self, Reference) and self.field is other.field and self.alias == other.alias

    def __hash__(self):
        return hash("reference{}{}".format(self.alias, self.field))

    def __repr__(self):
        return "{}({})".format(self.alias, self.field.alias)

    @immutable
    def for_(self, field):
        self.field = field


class ReferenceType(object):
    def __init__(self, alias: str, label: str, time_unit: str, interval: int):
        self.alias = alias
        self.label = label

        self.time_unit = time_unit
        self.interval = interval

    def __call__(
        self, dimension, delta: bool = False, delta_percent: bool = False, filters: List[ReferenceFilter] = ()
    ):
        return Reference(dimension, self, delta=delta, delta_percent=delta_percent, filters=filters)

    def __eq__(self, other):
        return isinstance(self, ReferenceType) and self.time_unit == other.time_unit and self.interval == other.interval

    def __hash__(self):
        return hash("reference" + self.alias)


def _over_format(identifier: str, interval: int):
    reference_format = f'{identifier}o{identifier}'

    if interval > 1:
        reference_format = f'{identifier}o{interval}{identifier}'

    return reference_format


DaysOverDays: Callable[[int], ReferenceType] = lambda interval: ReferenceType(
    _over_format('d', interval), _over_format('D', interval), 'day', interval
)
WeeksOverWeeks: Callable[[int], ReferenceType] = lambda interval: ReferenceType(
    _over_format('w', interval), _over_format('W', interval), 'week', interval
)
MonthsOverMonths: Callable[[int], ReferenceType] = lambda interval: ReferenceType(
    _over_format('m', interval), _over_format('M', interval), 'month', interval
)
QuartersOverQuarters: Callable[[int], ReferenceType] = lambda interval: ReferenceType(
    _over_format('q', interval), _over_format('Q', interval), 'quarter', interval
)
YearsOverYears: Callable[[int], ReferenceType] = lambda interval: ReferenceType(
    _over_format('y', interval), _over_format('Y', interval), 'year', interval
)

DayOverDay = DaysOverDays(interval=1)
WeekOverWeek = WeeksOverWeeks(interval=1)
MonthOverMonth = MonthsOverMonths(interval=1)
QuarterOverQuarter = QuartersOverQuarters(interval=1)
YearOverYear = YearsOverYears(interval=1)


def reference_term(reference: Reference, original_query: QueryBuilder, ref_query: QueryBuilder):
    """
    Part of query building. Given a reference, the original dataset query, and the ref query, creates the pypika for
    the reference that should be selected in the reference container query.

    :param reference:
    :param original_query:
    :param ref_query:
    :return:
    """

    def original_field(metric):
        return original_query.field(utils.alias_selector(metric.alias))

    def ref_field(metric):
        return ref_query.field(utils.alias_selector(metric.alias))

    if reference.delta:
        if reference.delta_percent:
            return lambda metric: (original_field(metric) - ref_field(metric)) * (100 / fn.NullIf(ref_field(metric), 0))

        return lambda metric: original_field(metric) - ref_field(metric)

    return ref_field


def calculate_delta_percent(ref_df, ref_delta_df):
    # pandas raises an exception when dividing by zero
    return 100.0 * ref_delta_df.divide(ref_df.replace(0, np.nan))
